@@ -1,75 +1,184 @@
-from BeautifulSoup import BeautifulStoneSoup
+from __future__ import absolute_import, with_statement
+
+import sys
 import decimal
 import datetime
 import codecs
-import mcc
 import re
+import collections
+import contextlib
+
+try:
+    from StringIO import StringIO
+except ImportError:
+    from io import StringIO
+
+import six
+
+if 'OrderedDict' in dir(collections):
+    odict = collections
+else:
+    import ordereddict as odict
+
+from . import mcc
+
+def soup_maker(fh):
+    try:
+        from bs4 import BeautifulSoup
+        return BeautifulSoup(fh)
+    except ImportError:
+        from BeautifulSoup import BeautifulStoneSoup
+        return BeautifulStoneSoup(fh)
+
+
+def try_decode(string, encoding):
+    if hasattr(string, 'decode'):
+        string = string.decode(encoding)
+    return string
+
+def is_iterable(candidate):
+    if sys.version_info < (2,6):
+        return hasattr(candidate, 'next')
+    return isinstance(candidate, collections.Iterable)
+
+@contextlib.contextmanager
+def save_pos(fh):
+    """
+    Save the position of the file handle, seek to the beginning, and
+    then restore the position.
+    """
+    orig_pos = fh.tell()
+    fh.seek(0)
+    try:
+        yield fh
+    finally:
+        fh.seek(orig_pos)
 
 class OfxFile(object):
     def __init__(self, fh):
-        self.headers = {}
+        """
+        fh should be a seekable file-like byte stream object
+        """
+        self.headers = odict.OrderedDict()
         self.fh = fh
-        self.read_headers()
 
-    def read_headers(self):
-        if not hasattr(self.fh, "seek") or not hasattr(self.fh, "next"):
+        if not is_iterable(self.fh):
+            return
+        if not hasattr(self.fh, "seek"):
             return  # fh is not a file object, we're doomed.
 
-        orig_pos = self.fh.tell()
-        self.fh.seek(0)
+        with save_pos(self.fh):
+            self.read_headers()
+            self.handle_encoding()
+            self.replace_NONE_headers()
 
+    def read_headers(self):
         head_data = self.fh.read(1024 * 10)
-        head_data = head_data[:head_data.find('<')]
+        head_data = head_data[:head_data.find(six.b('<'))]
 
-        for line in re.split('\r?\n?', head_data):
+        for line in re.split(six.b('\r?\n?'), head_data):
             # Newline?
-            if line.strip() == "":
+            if line.strip() == six.b(""):
                 break
 
-            header, value = line.split(":")
+            header, value = line.split(six.b(":"))
             header, value = header.strip().upper(), value.strip()
 
-            if value.upper() == "NONE":
-                value = None
-
             self.headers[header] = value
 
-        # Look for the encoding
-        enc_type = self.headers.get("ENCODING")
-        if enc_type:
-            encoding = None  # Unknown
+    def handle_encoding(self):
+        """
+        Decode the headers and wrap self.fh in a decoder such that it
+        subsequently returns only text.
+        """
+        # decode the headers using ascii
+        ascii_headers = odict.OrderedDict(
+            (
+                key.decode('ascii', 'replace'),
+                value.decode('ascii', 'replace'),
+            )
+            for key, value in six.iteritems(self.headers)
+        )
 
-            if enc_type == "USASCII":
-                cp = self.headers.get("CHARSET", "1252")
-                encoding = "cp%s" % (cp, )
+        enc_type = ascii_headers.get('ENCODING')
 
-            elif enc_type in ("UNICODE", "UTF-8"):
-                encoding = "utf-8"
+        if not enc_type:
+            # no encoding specified, use the ascii-decoded headers
+            self.headers = ascii_headers
+            # decode the body as ascii as well
+            self.fh = codecs.lookup('ascii').streamreader(self.fh)
+            return
 
-            try:
-                codec = codecs.lookup(encoding)
-            except LookupError:
-                encoding = None
+        if enc_type == "USASCII":
+            cp = ascii_headers.get("CHARSET", "1252")
+            encoding = "cp%s" % (cp, )
 
-            if encoding:
-                self.fh = codec.streamreader(self.fh)
+        elif enc_type in ("UNICODE", "UTF-8"):
+            encoding = "utf-8"
 
-                # Decode the headers
-                uheaders = {}
-                for key, value in self.headers.iteritems():
-                    key = key.decode(encoding)
+        codec = codecs.lookup(encoding)
 
-                    if type(value) is str:
-                        value = value.decode(encoding)
+        self.fh = codec.streamreader(self.fh)
 
-                    uheaders[key] = value
-                self.headers = uheaders
-        # Reset the fh to the original position
-        self.fh.seek(orig_pos)
+        # Decode the headers using the encoding
+        self.headers = odict.OrderedDict(
+            (key.decode(encoding), value.decode(encoding))
+            for key, value in six.iteritems(self.headers)
+        )
+
+    def replace_NONE_headers(self):
+        """
+        Any headers that indicate 'none' should be replaced with Python
+        None values
+        """
+        for header in self.headers:
+            if self.headers[header].upper() == 'NONE':
+                self.headers[header] = None
+
+
+class OfxPreprocessedFile(OfxFile):
+    def __init__(self, fh):
+        super(OfxPreprocessedFile,self).__init__(fh)
+
+        if self.fh is None:
+            return
+
+        ofx_string = self.fh.read()
+
+        # find all closing tags as hints
+        closing_tags = [ t.upper() for t in re.findall(r'(?i)</([a-z0-9_\.]+)>', ofx_string) ]
+
+        # close all tags that don't have closing tags and
+        # leave all other data intact
+        last_open_tag = None
+        tokens        = re.split(r'(?i)(</?[a-z0-9_\.]+>)', ofx_string)
+        new_fh        = StringIO()
+        for idx,token in enumerate(tokens):
+            is_closing_tag = token.startswith('</')
+            is_processing_tag = token.startswith('<?')
+            is_cdata = token.startswith('<!')
+            is_tag = token.startswith('<') and not is_cdata
+            is_open_tag = is_tag and not is_closing_tag and not is_processing_tag
+            if is_tag:
+                if last_open_tag is not None:
+                    new_fh.write("</%s>" % last_open_tag)
+                    last_open_tag = None
+            if is_open_tag:
+                tag_name = re.findall(r'(?i)<([a-z0-9_\.]+)>', token)[0]
+                if tag_name.upper() not in closing_tags:
+                    last_open_tag = tag_name
+            new_fh.write(token)
+        new_fh.seek(0)
+        self.fh = new_fh
 
 
 class Ofx(object):
-    pass
+    def __str__(self):
+        return ""
+#        headers = "\r\n".join(":".join(el if el else "NONE" for el in item) for item in six.iteritems(self.headers))
+#        headers += "\r\n\r\n"
+#
+#        return headers + str(self.signon)
 
 
 class AccountType(object):
@@ -81,6 +190,7 @@ class Account(object):
         self.statement = None
         self.account_id = ''
         self.routing_number = ''
+        self.branch_id = ''
         self.account_type = ''
         self.institution = None
         self.type = AccountType.Unknown
@@ -106,6 +216,24 @@ class Security:
         self.ticker = ticker
         self.memo = memo
 
+class Signon:
+    def __init__(self, code, severity, message):
+        self.code = code
+        self.severity = severity
+        self.message = message
+        if int(code) == 0:
+            self.success = True
+        else:
+            self.success = False
+
+    def __str__(self):
+        ret = "\t<SIGNONMSGSRSV1>\r\n" + "\t\t<SONRS>\r\n" + "\t\t\t<STATUS>\r\n"
+        ret += "\t\t\t\t<CODE>%s\r\n" % self.code
+        ret += "\t\t\t\t<SEVERITY>%s\r\n" % self.severity
+        if self.message:
+            ret += "\t\t\t\t<MESSAGE>%s\r\n" % self.message
+        ret += "\t\t\t</STATUS>\r\n" + "\t\t</SONRS>\r\n" + "\t</SIGNONMSGSRSV1>\r\n"
+        return ret
 
 class Statement(object):
     def __init__(self):
@@ -137,13 +265,19 @@ class Transaction(object):
         self.memo = ''
         self.sic = None
         self.mcc = ''
+        self.checknum = ''
 
     def __repr__(self):
         return "<Transaction units=" + str(self.amount) + ">"
 
 
 class InvestmentTransaction(object):
-    def __init__(self):
+    (Unknown, BuyMF, SellMF, Reinvest, BuyStock, SellStock) = [x for x in range(-1, 5)]
+    def __init__(self, type):
+        try:
+            self.type = ['buymf', 'sellmf', 'reinvest', 'buystock', 'sellstock'].index(type.lower())
+        except ValueError:
+            self.type = InvestmentTransaction.Unknown
         self.tradeDate = None
         self.settleDate = None
         self.security = ''
@@ -151,7 +285,7 @@ class InvestmentTransaction(object):
         self.unit_price = decimal.Decimal(0)
 
     def __repr__(self):
-        return "<InvestmentTransaction units=" + str(self.units) + ">"
+        return "<InvestmentTransaction type=" + str(self.type) + ", units=" + str(self.units) + ">"
 
 
 class Position(object):
@@ -188,19 +322,24 @@ class OfxParser(object):
         cls_.fail_fast = fail_fast
 
         if isinstance(file_handle, type('')):
-            raise RuntimeError(u"parse() takes in a file handle, not a string")
+            raise RuntimeError(six.u("parse() takes in a file handle, not a string"))
 
         ofx_obj = Ofx()
 
         # Store the headers
-        ofx_file = OfxFile(file_handle)
+        ofx_file = OfxPreprocessedFile(file_handle)
         ofx_obj.headers = ofx_file.headers
         ofx_obj.accounts = []
+        ofx_obj.signon = None
 
-        ofx = BeautifulStoneSoup(ofx_file.fh)
+        ofx = soup_maker(ofx_file.fh)
         if len(ofx.contents) == 0:
             raise OfxParserException('The ofx file is empty!')
 
+        sonrs_ofx = ofx.find('sonrs')
+        if sonrs_ofx:
+            ofx_obj.signon = cls_.parseSonrs(sonrs_ofx)
+
         stmtrs_ofx = ofx.findAll('stmtrs')
         if stmtrs_ofx:
             ofx_obj.accounts += cls_.parseStmtrs(stmtrs_ofx, AccountType.Bank)
@@ -270,10 +409,11 @@ class OfxParser(object):
             else:
                 continue
 
-	        fi_ofx = ofx.find('fi')
-	        if fi_ofx:
-	            for account in ofx_obj.accounts:
-	                account.institution = cls_.parseOrg(fi_ofx)
+                fi_ofx = ofx.find('fi')
+                if fi_ofx:
+                    for account in ofx_obj.accounts:
+                        account.institution = cls_.parseOrg(fi_ofx)
+
             desc = i.find('desc')
             if hasattr(desc, 'contents'):
                 for account in accounts:
@@ -292,7 +432,7 @@ class OfxParser(object):
                     account.account_id = acctid_tag.contents[0].strip()
                 except IndexError:
                     account.warnings.append(
-                        u"Empty acctid tag for %s" % invstmtrs_ofx)
+                        six.u("Empty acctid tag for %s") % invstmtrs_ofx)
                     if cls_.fail_fast:
                         raise
 
@@ -302,7 +442,7 @@ class OfxParser(object):
                     account.brokerid = brokerid_tag.contents[0].strip()
                 except IndexError:
                     account.warnings.append(
-                        u"Empty brokerid tag for %s" % invstmtrs_ofx)
+                        six.u("Empty brokerid tag for %s") % invstmtrs_ofx)
                     if cls_.fail_fast:
                         raise
 
@@ -357,7 +497,7 @@ class OfxParser(object):
 
     @classmethod
     def parseInvestmentTransaction(cls_, ofx):
-        transaction = InvestmentTransaction()
+        transaction = InvestmentTransaction(ofx.name)
         tag = ofx.find('fitid')
         if (hasattr(tag, 'contents')):
             transaction.id = tag.contents[0].strip()
@@ -403,11 +543,12 @@ class OfxParser(object):
                     statement.start_date = cls_.parseOfxDateTime(
                         tag.contents[0].strip())
                 except IndexError:
-                    statement.warnings.append(u'Empty start date.')
+                    statement.warnings.append(six.u('Empty start date.'))
                     if cls_.fail_fast:
                         raise
-                except ValueError, e:
-                    statement.warnings.append(u'Invalid start date: %s' % e)
+                except ValueError:
+                    e = sys.exc_info()[1]
+                    statement.warnings.append(six.u('Invalid start date: %s') % e)
                     if cls_.fail_fast:
                         raise
 
@@ -417,38 +558,41 @@ class OfxParser(object):
                     statement.end_date = cls_.parseOfxDateTime(
                         tag.contents[0].strip())
                 except IndexError:
-                    statement.warnings.append(u'Empty end date.')
-                except ValueError, e:
-                    statement.warnings.append(u'Invalid end date: %s' % e)
+                    statement.warnings.append(six.u('Empty end date.'))
+                except ValueError:
+                    e = sys.exc_info()[1]
+                    statement.warnings.append(six.u('Invalid end date: %s') % e)
                     if cls_.fail_fast:
                         raise
 
-        for transaction_type in ['posmf', 'posstock']:
+        for transaction_type in ['posmf', 'posstock', 'posopt']:
             try:
                 for investment_ofx in invstmtrs_ofx.findAll(transaction_type):
                     statement.positions.append(
                         cls_.parseInvestmentPosition(investment_ofx))
             except (ValueError, IndexError, decimal.InvalidOperation,
-                    TypeError), e:
+                    TypeError):
+                e = sys.exc_info()[1]
                 if cls_.fail_fast:
                     raise
                 statement.discarded_entries.append(
-                    {u'error': u"Error parsing positions: " + str(e),
-                     u'content': investment_ofx}
+                    {six.u('error'): six.u("Error parsing positions: ") + str(e),
+                     six.u('content'): investment_ofx}
                 )
 
         for transaction_type in ['buymf', 'sellmf', 'reinvest', 'buystock',
-                                 'sellstock']:
+                                 'sellstock', 'buyopt', 'sellopt']:
             try:
                 for investment_ofx in invstmtrs_ofx.findAll(transaction_type):
                     statement.transactions.append(
                         cls_.parseInvestmentTransaction(investment_ofx))
-            except (ValueError, IndexError, decimal.InvalidOperation), e:
+            except (ValueError, IndexError, decimal.InvalidOperation):
+                e = sys.exc_info()[1]
                 if cls_.fail_fast:
                     raise
                 statement.discarded_entries.append(
-                    {u'error': transaction_type + ": " + str(e),
-                     u'content': investment_ofx}
+                    {six.u('error'): transaction_type + ": " + str(e),
+                     six.u('content'): investment_ofx}
                 )
 
         return statement
@@ -466,6 +610,18 @@ class OfxParser(object):
 
         return institution
 
+    @classmethod
+    def parseSonrs(cls_, sonrs):
+
+        code     = int(sonrs.find('code').contents[0].strip())
+        severity = sonrs.find('severity').contents[0].strip()
+        try:
+            message = sonrs.find('message').contents[0].strip()
+        except:
+            message = ''
+
+        return Signon(code,severity,message)
+
     @classmethod
     def parseStmtrs(cls_, stmtrs_list, accountType):
         ''' Parse the <STMTRS> tags and return a list of Accounts object. '''
@@ -478,6 +634,9 @@ class OfxParser(object):
             bankid_tag = stmtrs_ofx.find('bankid')
             if hasattr(bankid_tag, 'contents'):
                 account.routing_number = bankid_tag.contents[0].strip()
+            branchid_tag = stmtrs_ofx.find('branchid')
+            if hasattr(branchid_tag, 'contents'):
+                account.branch_id = branchid_tag.contents[0].strip()
             type_tag = stmtrs_ofx.find('accttype')
             if hasattr(type_tag, 'contents'):
                 account.account_type = type_tag.contents[0].strip()
@@ -501,12 +660,12 @@ class OfxParser(object):
                     dtstart_tag.contents[0].strip())
             except IndexError:
                 statement.warnings.append(
-                    u"Statement start date was empty for %s" % stmt_ofx)
+                    six.u("Statement start date was empty for %s") % stmt_ofx)
                 if cls_.fail_fast:
                     raise
             except ValueError:
                 statement.warnings.append(
-                    u"Statement start date was not allowed for %s" % stmt_ofx)
+                    six.u("Statement start date was not allowed for %s") % stmt_ofx)
                 if cls_.fail_fast:
                     raise
 
@@ -517,18 +676,19 @@ class OfxParser(object):
                     dtend_tag.contents[0].strip())
             except IndexError:
                 statement.warnings.append(
-                    u"Statement start date was empty for %s" % stmt_ofx)
+                    six.u("Statement start date was empty for %s") % stmt_ofx)
                 if cls_.fail_fast:
                     raise
-            except ValueError, ve:
-                statement.warnings.append(
-                    u"Statement start date was not formatted correctly for"
-                    u" %s" % stmt_ofx)
+            except ValueError:
+                ve = sys.exc_info()[1]
+                msg = six.u("Statement start date was not formatted "
+                    "correctly for %s")
+                statement.warnings.append(msg % stmt_ofx)
                 if cls_.fail_fast:
                     raise
             except TypeError:
                 statement.warnings.append(
-                    u"Statement start date was not allowed for %s" % stmt_ofx)
+                    six.u("Statement start date was not allowed for %s") % stmt_ofx)
                 if cls_.fail_fast:
                     raise
 
@@ -538,7 +698,7 @@ class OfxParser(object):
                 statement.currency = currency_tag.contents[0].strip().lower()
             except IndexError:
                 statement.warnings.append(
-                    u"Currency definition was empty for %s" % stmt_ofx)
+                    six.u("Currency definition was empty for %s") % stmt_ofx)
                 if cls_.fail_fast:
                     raise
 
@@ -549,9 +709,10 @@ class OfxParser(object):
                 try:
                     statement.balance = decimal.Decimal(
                         balamt_tag.contents[0].strip())
-                except (IndexError, decimal.InvalidOperation), ex:
+                except (IndexError, decimal.InvalidOperation):
+                    ex = sys.exc_info()[1]
                     statement.warnings.append(
-                        u"Ledger balance amount was empty for %s" % stmt_ofx)
+                        six.u("Ledger balance amount was empty for %s") % stmt_ofx)
                     if cls_.fail_fast:
                         raise OfxParserException("Empty ledger balance")
 
@@ -562,9 +723,10 @@ class OfxParser(object):
                 try:
                     statement.available_balance = decimal.Decimal(
                         balamt_tag.contents[0].strip())
-                except (IndexError, decimal.InvalidOperation), ex:
-                    statement.warnings.append(u"Available balance amount was"
-                                              u" empty for %s" % stmt_ofx)
+                except (IndexError, decimal.InvalidOperation):
+                    ex = sys.exc_info()[1]
+                    msg = six.u("Available balance amount was empty for %s")
+                    statement.warnings.append(msg % stmt_ofx)
                     if cls_.fail_fast:
                         raise OfxParserException("Empty available balance")
 
@@ -572,7 +734,8 @@ class OfxParser(object):
             try:
                 statement.transactions.append(
                     cls_.parseTransaction(transaction_ofx))
-            except OfxParserException, ofxError:
+            except OfxParserException:
+                ofxError = sys.exc_info()[1]
                 statement.discarded_entries.append(
                     {'error': str(ofxError), 'content': transaction_ofx})
                 if cls_.fail_fast:
@@ -592,20 +755,20 @@ class OfxParser(object):
             try:
                 transaction.type = type_tag.contents[0].lower().strip()
             except IndexError:
-                raise OfxParserException(u"Empty transaction type")
+                raise OfxParserException(six.u("Empty transaction type"))
             except TypeError:
                 raise OfxParserException(
-                    u"No Transaction type (a required field)")
+                    six.u("No Transaction type (a required field)"))
 
         name_tag = txn_ofx.find('name')
         if hasattr(name_tag, "contents"):
             try:
                 transaction.payee = name_tag.contents[0].strip()
             except IndexError:
-                raise OfxParserException(u"Empty transaction name")
+                raise OfxParserException(six.u("Empty transaction name"))
             except TypeError:
                 raise OfxParserException(
-                    u"No Transaction name (a required field)")
+                    six.u("No Transaction name (a required field)"))
 
         memo_tag = txn_ofx.find('memo')
         if hasattr(memo_tag, "contents"):
@@ -626,13 +789,13 @@ class OfxParser(object):
                 raise OfxParserException("Invalid Transaction Date")
             except decimal.InvalidOperation:
                 raise OfxParserException(
-                    u"Invalid Transaction Amount: '%s'" % amt_tag.contents[0])
+                    six.u("Invalid Transaction Amount: '%s'") % amt_tag.contents[0])
             except TypeError:
                 raise OfxParserException(
-                    u"No Transaction Amount (a required field)")
+                    six.u("No Transaction Amount (a required field)"))
         else:
             raise OfxParserException(
-                u"Missing Transaction Amount (a required field)")
+                six.u("Missing Transaction Amount (a required field)"))
 
         date_tag = txn_ofx.find('dtposted')
         if hasattr(date_tag, "contents"):
@@ -641,38 +804,45 @@ class OfxParser(object):
                     date_tag.contents[0].strip())
             except IndexError:
                 raise OfxParserException("Invalid Transaction Date")
-            except ValueError, ve:
+            except ValueError:
+                ve = sys.exc_info()[1]
                 raise OfxParserException(str(ve))
             except TypeError:
                 raise OfxParserException(
-                    u"No Transaction Date (a required field)")
+                    six.u("No Transaction Date (a required field)"))
         else:
             raise OfxParserException(
-                u"Missing Transaction Date (a required field)")
+                six.u("Missing Transaction Date (a required field)"))
 
         id_tag = txn_ofx.find('fitid')
         if hasattr(id_tag, "contents"):
             try:
                 transaction.id = id_tag.contents[0].strip()
             except IndexError:
-                raise OfxParserException(u"Empty FIT id (a required field)")
+                raise OfxParserException(six.u("Empty FIT id (a required field)"))
             except TypeError:
-                raise OfxParserException(u"No FIT id (a required field)")
+                raise OfxParserException(six.u("No FIT id (a required field)"))
         else:
-            raise OfxParserException(u"Missing FIT id (a required field)")
+            raise OfxParserException(six.u("Missing FIT id (a required field)"))
 
         sic_tag = txn_ofx.find('sic')
         if hasattr(sic_tag, 'contents'):
             try:
-				transaction.sic = sic_tag.contents[0].strip()
+                transaction.sic = sic_tag.contents[0].strip()
             except IndexError:
-                raise OfxParserException(u"Empty transaction Standard Industry Code (SIC)")
+                raise OfxParserException(six.u("Empty transaction Standard Industry Code (SIC)"))
 
         if transaction.sic is not None:
             try:
                 transaction.mcc = mcc.codes.get(transaction.sic, '').get('combined description')
             except IndexError:
-                raise OfxParserException(u"Empty transaction Merchant Category Code (MCC)")
+                raise OfxParserException(six.u("Empty transaction Merchant Category Code (MCC)"))
 
+        checknum_tag = txn_ofx.find('checknum')
+        if hasattr(checknum_tag, 'contents'):
+            try:
+                transaction.checknum = checknum_tag.contents[0].strip()
+            except IndexError:
+                raise OfxParserException(six.u("Empty Check (or other reference) number"))
 
         return transaction
